Package: sesameData
Type: Package
Title: Supporting Data for SeSAMe Package
Description: Provides supporting annotation and test data for SeSAMe package. This includes chip tango addresses, mapping information, performance annotation, and trained predictor for Infinium array data. This package provides user access to essential annotation data for working with many generations of the Infinium DNA methylation array. Current we support human array (HM27, HM450, EPIC), mouse array (MM285) and the HorvathMethylChip40 (Mammal40) array.
Version: 1.28.0
Authors@R: c(person("Wanding", "Zhou", role = c("aut","cre"),
             email = "zhouwanding@gmail.com", comment = c(ORCID = "0000-0001-9126-1932")),
             person("Hui", "Shen", role = c("aut"),
             email = "Hui.Shen@vai.org"),
             person("Timothy", "Triche", role = "ctb"))
License: Artistic-2.0
Depends: R (>= 4.2.0), ExperimentHub, AnnotationHub
Imports: utils, readr, stringr, GenomicRanges (>= 1.61.1), S4Vectors,
        IRanges, Seqinfo
Suggests: BiocGenerics, GenomeInfoDb, sesame, testthat, knitr,
        rmarkdown
biocViews: ExperimentData, MicroarrayData, Genome, ExperimentHub,
        MethylationArrayData
URL: https://github.com/zwdzwd/sesame
BugReports: https://github.com/zwdzwd/sesame/issue
VignetteBuilder: knitr
NeedsCompilation: no
RoxygenNote: 7.3.2
Encoding: UTF-8
git_url: https://git.bioconductor.org/packages/sesameData
git_branch: RELEASE_3_22
git_last_commit: 851657b
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.22
Date/Publication: 2025-11-04
Packaged: 2025-11-04 16:43:18 UTC; biocbuild
Author: Wanding Zhou [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-9126-1932>),
  Hui Shen [aut],
  Timothy Triche [ctb]
Maintainer: Wanding Zhou <zhouwanding@gmail.com>

# Generated by roxygen2: do not edit by hand

export(inferPlatformFromProbeIDs)
export(sesameDataCache)
export(sesameDataCacheAll)
export(sesameDataGet)
export(sesameDataGet_resetEnv)
export(sesameDataHas)
export(sesameDataList)
export(sesameData_annoProbes)
export(sesameData_check_genome)
export(sesameData_check_platform)
export(sesameData_getGenomeInfo)
export(sesameData_getManifestGRanges)
export(sesameData_getProbesByGene)
export(sesameData_getProbesByRegion)
export(sesameData_getTxnGRanges)
export(sesameData_txnToGeneGRanges)
import(AnnotationHub)
import(ExperimentHub)
import(GenomicRanges)
import(stringr)
import(utils)
importFrom(GenomicRanges,end)
importFrom(GenomicRanges,findOverlaps)
importFrom(GenomicRanges,mcols)
importFrom(GenomicRanges,start)
importFrom(GenomicRanges,trim)
importFrom(IRanges,IRanges)
importFrom(S4Vectors,queryHits)
importFrom(S4Vectors,subjectHits)
importFrom(Seqinfo,Seqinfo)
importFrom(readr,col_character)
importFrom(readr,col_integer)
importFrom(readr,cols)
importFrom(readr,read_tsv)
importFrom(stringr,str_replace)
importMethodsFrom(IRanges,subsetByOverlaps)

# SeSAMe Supporting Data Package

[![last commit](https://img.shields.io/github/last-commit/zwdzwd/sesameData.svg?style=flat-square)](https://github.com/zwdzwd/sesameData/commits/master)
                  
## Bugs
    
Bug reports are appreciated. Register issues at the SeSAMe [issue tracker](http://github.com/zwdzwd/sesame/issues)
    
    
## About

See [SeSAMe: reducing artifactual detection of DNA methylation by Infinium BeadChips in genomic deletions](https://doi.org/10.1093/nar/gky691) for more details.

## Exported functions (user-facing)

Data access and caching:
- `sesameDataList()` list available data titles
- `sesameDataHas()` check if titles exist
- `sesameDataGet()` load a data object by title
- `sesameDataCache()` / `sesameDataCacheAll()` cache data locally
- `sesameDataGet_resetEnv()` clear the in-memory cache

Platform and genome helpers:
- `inferPlatformFromProbeIDs()` infer platform from probe IDs
- `sesameData_check_platform()` validate or infer platform
- `sesameData_check_genome()` validate or infer genome build
- `sesameData_getGenomeInfo()` load genome metadata

Manifests, probes, and annotation:
- `sesameData_getManifestGRanges()` load manifest GRanges
- `sesameData_getProbesByRegion()` select probes by genomic range
- `sesameData_getProbesByGene()` select probes by gene/promoter
- `sesameData_annoProbes()` annotate probes by features

Transcript utilities:
- `sesameData_getTxnGRanges()` get transcripts or merged genes
- `sesameData_txnToGeneGRanges()` merge transcripts to genes

## Helper function to extract labels from overlaps
## Extracts annotation labels from overlapping regions, either using region
## names or a specific metadata column. Returns both the labels and the
## output column name to use.
.getLabelsFromOverlaps <- function(regs, hits, column, out_name) {
    if (is.null(column)) {
        # Use region names as labels when no column specified
        label <- names(regs[subjectHits(hits)])
        if (is.null(out_name)) { out_name <- "anno" }
    } else {
        # Use specified metadata column for labels
        stopifnot(column %in% colnames(mcols(regs)))
        label <- mcols(regs[subjectHits(hits)])[[column]]
        if (is.null(out_name)) { out_name <- column }
    }
    list(label = label, out_name = out_name)
}

## Helper function to collapse annotations
## When a probe overlaps multiple regions, this function either picks the
## first annotation (chooseOne=TRUE) or concatenates all unique annotations
## with a separator. Probes without overlaps get NA.
.collapseAnnotations <- function(
    probes, hits, label, out_name, chooseOne, sep) {
    if (chooseOne) {
        # Take only the first annotation for each probe
        pid2label <- vapply(split(label, queryHits(hits)),
            function(x) x[1], character(1))
    } else {
        # Concatenate all unique annotations with separator
        pid2label <- vapply(split(label, queryHits(hits)),
            function(x) paste0(unique(x), collapse=sep), character(1))
    }
    # Initialize all probes with NA, then fill in the annotations
    mcols(probes)[[out_name]] <- NA
    mcols(probes[as.integer(names(pid2label))])[[out_name]] <- pid2label
    probes
}

#' Annotate Probes by Probe ID
#'
#' This function annotates probes based on genomic coordinate overlaps with
#' provided genomic regions (GRanges). Columns in the manifests will be added
#' to the annotation. Please note that if unfound, the annotation will be NA.
#' The probe will always be kept in the output.
#'
#' For annotation by probe ID using KYCG databases (rather than by genomic
#' coordinates), see \code{knowYourCG::annoProbes()}
#'
#' @param Probe_IDs a character vector of probe IDs
#' @param regs a GenomicRanges::GRanges object against which
#' probes will be annotated, default to genes if not given
#' @param collapse whether to collapse multiple regs into one
#' @param chooseOne choose an arbitrary annotation if multiple exist
#' default to FALSE. which concatenates all with ","
#' @param sep the delimiter for collapsing
#' @param column which column in regs to annotate, if not given
#' return all overlapping probes
#' @param return_ov_probes if TRUE will return overlapping
#' probes in a GRanges object.
#' @param return_ov_features if TRUE will return overlapping
#' features in a GRanges object.
#' @param out_name column header of the annotation, use column if not given
#' @param platform EPIC, MM285 etc. will infer from Probe_IDs if not given
#' @param genome hg38, mm10, ... will infer if not given.
#' For additional mapping, download the GRanges object from
#' http://zwdzwd.github.io/InfiniumAnnotation
#' and provide the following argument
#' ..., genome = sesameAnno_buildManifestGRanges("downloaded_file"),...
#' to this function.
#' @param silent suppress messages
#' @importFrom GenomicRanges findOverlaps
#' @importFrom GenomicRanges mcols
#' @importFrom S4Vectors subjectHits
#' @importFrom S4Vectors queryHits
#' @return a GRanges with annotated column
#' If a probe has no overlap with regs, it will be included in the results
#' with NA. But if a probe is not included in the manifest (due to
#' mappability), it won't be included in the results.
#' 
#' @examples
#' library(GenomicRanges)
#' sesameDataCache(c(
#'     "genomeInfo.mm10", "MM285.address",
#'     "genomeInfo.hg38", "Mammal40.address"))
#'
#' ## Example 1: Basic usage - annotate with gene names (default)
#' ## When regs=NULL, function defaults to gene annotation
#' probes <- c("cg14620903", "cg22464003")
#' anno <- sesameData_annoProbes(probes)
#' ## Returns GRanges with gene_name column
#'
#' ## Example 2: Annotate mouse probes with promoter regions
#' regs <- sesameData_getTxnGRanges("mm10")
#' Probe_IDs <- names(sesameData_getManifestGRanges("MM285"))[1:100]
#' anno <- sesameData_annoProbes(Probe_IDs, promoters(regs), column="gene_name")
#' ## Probes overlapping promoters are annotated with gene names
#'
#' ## Example 3: Get features that overlap with probes
#' ## return_ov_features=TRUE returns the overlapping genomic features
#' genes <- sesameData_getTxnGRanges("hg38", merge2gene = TRUE)
#' ov_genes <- sesameData_annoProbes(
#'     c("cg14620903","cg22464003"), genes, return_ov_features=TRUE)
#' ## Returns GRanges of genes overlapping the probes
#'
#' ## Example 4: Get probes that overlap with features
#' ## return_ov_probes=TRUE returns only overlapping probes
#' ov_probes <- sesameData_annoProbes(
#'     c("cg14620903","cg22464003"), genes, return_ov_probes=TRUE)
#' ## Returns GRanges of probes that overlap genes
#'
#' ## Example 5: Control annotation collapsing behavior
#' ## collapse=TRUE (default): multiple annotations concatenated with separator
#' anno_collapsed <- sesameData_annoProbes(
#'     c("cg14620903","cg22464003"), genes, column="gene_name",
#'     collapse=TRUE, sep=";")
#' ## Multiple overlapping genes listed as "GENE1;GENE2;GENE3"
#'
#' ## collapse=FALSE: each probe-feature overlap gets separate entry
#' anno_expanded <- sesameData_annoProbes(
#'     c("cg14620903","cg22464003"), genes, column="gene_name",
#'     collapse=FALSE)
#' ## Probes with multiple overlaps appear multiple times
#'
#' ## Example 6: Choose only first annotation when multiple exist
#' anno_one <- sesameData_annoProbes(
#'     c("cg14620903","cg22464003"), genes, column="gene_name",
#'     chooseOne=TRUE)
#' ## Each probe gets only the first overlapping gene
#'
#' ## Example 7: Annotate with custom genomic regions
#' custom_regs <- GRanges(
#'     seqnames = c("chr5", "chr5"),
#'     ranges = IRanges(start = c(10000, 135350870),
#'                      end = c(135350866, 145369531)),
#'     feature_type = c("enhancer", "silencer"))
#' anno_custom <- sesameData_annoProbes(
#'     c("cg14620903","cg22464003"), custom_regs,
#'     column="feature_type", genome="hg38")
#'
#' ## Note: For annotation by probe ID using KYCG databases
#' ## (rather than genomic coordinates), see knowYourCG::annoProbes()
#' @export
sesameData_annoProbes <- function(Probe_IDs, regs = NULL,
    collapse = TRUE, chooseOne = FALSE, column = NULL, sep=",",
    return_ov_probes = FALSE, return_ov_features = FALSE,
    out_name = NULL, platform = NULL, genome = NULL, silent = FALSE) {

    # Validate input
    stopifnot(is.character(Probe_IDs))

    # Infer platform from probe IDs if not provided
    if(is.null(platform)) {
        platform <- inferPlatformFromProbeIDs(Probe_IDs, silent = silent)
    }

    # Set up default genomic regions (genes) if not provided
    if (is.null(regs)) {
        if (is.null(genome)) {
            genome <- sesameData_check_genome(NULL, platform)
        }
        regs <- sesameData_getTxnGRanges(genome)
        if (is.null(column)) column <- "gene_name"
    }

    # Get probe coordinates from manifest
    gr <- sesameData_getManifestGRanges(platform, genome = genome)
    in_mft <- Probe_IDs %in% names(gr)
    if (sum(!in_mft) > 0) {
        warning(sprintf(
            "%d probes out of manifest were excluded.", sum(!in_mft)))
    }

    # Subset to requested probes and find overlaps with regions
    probes <- gr[Probe_IDs[in_mft]]
    if (length(probes) == 0) { return(probes); }
    hits <- findOverlaps(probes, regs, ignore.strand = TRUE)

    # Return early if only requesting overlapping probes or features
    if (return_ov_probes) {
        return(probes[unique(queryHits(hits))])
    } else if (return_ov_features) {
        return(regs[unique(subjectHits(hits))])
    }

    # Extract labels from overlapping regions
    label_info <- .getLabelsFromOverlaps(regs, hits, column, out_name)
    label <- label_info$label
    out_name <- label_info$out_name

    # Annotate probes with labels
    if (collapse) {
        # Collapse multiple annotations per probe into single entries
        probes <- .collapseAnnotations(
            probes, hits, label, out_name, chooseOne, sep)
    } else {
        # Keep all probe-region pairs separate (probes may be duplicated)
        unfound <- probes[!(seq_along(probes) %in% queryHits(hits))]
        if (length(unfound) > 0) { mcols(unfound)[[out_name]] <- NA }
        found <- probes[queryHits(hits)]
        if (length(found) > 0) { mcols(found)[[out_name]] <- label }
        probes <- sort(c(unfound, found))
    }
    probes
}

sesameDataRecache <- function(eh_id) {
    stopifnot(startsWith(eh_id, "EH"))
    tryCatch({
        sesameDataCache0(eh_id)
    }, error=function(cond) {
        stopAndCache(eh_id)
    })
    query(ExperimentHub(localHub=TRUE), 'sesameData')
}

stopAndCache <- function(title) {
    stop(sprintf('
| File %s either not found or needs to be cached to be
| used in sesame.
| Please make sure you have updated ExperimentHub and try
| > sesameDataCache("%s")
| or download all data
| > sesameDataCache()
| to retrieve and cache needed sesame data.', title, title))
}

sesameDataCache0 <- function(eh_ids) {
    ## load meta data
    message(sprintf("Metadata (N=%d):\n", length(eh_ids)))
    suppressMessages(log <- capture.output(
        eh <- query(ExperimentHub(), "sesameData")[eh_ids]))
    
    ## load actual data
    tmp2 <- lapply(seq_along(eh), function(i) {
        message(sprintf(
            "(%d/%d) %s:\n", i, length(eh_ids), eh_ids[i]))
        suppressMessages(log <- capture.output(cache(eh[i])))
    })
}

#' Cache SeSAMe data
#'
#' @import ExperimentHub
#' @import AnnotationHub
#' @param data_titles data to cache, if not given will cache all
#' @return TRUE
#' @examples
#' sesameDataCache("genomeInfo.hg38")
#' @export
sesameDataCache <- function(data_titles = NULL) {
    setExperimentHubOption(arg="MAX_DOWNLOADS", 100)
    if (is.null(data_titles)) { # cache all
        eh_ids <- unique(df_master$EHID)
    } else if (all(data_titles %in% df_master$Title)) {
        eh_ids <- df_master$EHID[df_master$Title %in% data_titles]
    } else {
        data_titles <- data_titles[!(data_titles %in% df_master$Title)]
        stop(sprintf("%s is missing from sesameData.", data_titles[1]))
    }
    eh_ids <- eh_ids[eh_ids != "TBD"]
    eh_ids <- eh_ids[!is.na(eh_ids)]
    stopifnot(length(eh_ids) > 0)
    suppressMessages(try({
        eh_ids <- eh_ids[!(eh_ids %in% names(ExperimentHub(localHub=TRUE)))]
    }, silent = TRUE))
    if (length(eh_ids) == 0) return(invisible(TRUE));
    tryCatch({
        sesameDataCache0(eh_ids)
    }, error = function(cond) {
        message("ExperimentHub Caching fails:")
        message(cond)
        return(invisible(FALSE))
    })
    invisible(TRUE)
}

#' Cache all SeSAMe data
#' 
#' @param data_titles data to cache, if not given will cache all
#' @return TRUE
#' @import ExperimentHub
#' @import AnnotationHub
#' @examples
#' sesameDataCache("genomeInfo.hg38")
#' @export
sesameDataCacheAll <- sesameDataCache

#' infer platform from Probe_IDs
#'
#' @param Probe_IDs probe IDs
#' @param silent suppress message
#' @return a platform code
#' @examples
#' sesameDataCache("probeIDSignature")
#' inferPlatformFromProbeIDs(c("cg14620903","cg22464003"))
#' @export
inferPlatformFromProbeIDs <- function(Probe_IDs, silent = FALSE) {
    sig <- sesameDataGet("probeIDSignature")
    cnts <- vapply(sig, function(x) sum(Probe_IDs %in% x), integer(1))
    if(sum(cnts == max(cnts)) > 1) {
        stop("Ambiguous platform. Please provide platform explicitly.") }

    platform <- names(which.max(cnts))
    if (!silent) {
        message("Platform set to: ", platform)
    }
    platform
}

#' Check platform code
#'
#' Note: custome platforms lead to error here.
#' 
#' @param platform input platform
#' @param probes probes by which the platform may be guessed
#' @param silent suppress message
#' @return platform code
#' @examples
#' sesameData_check_platform("HM450")
#' @export
sesameData_check_platform <- function(
    platform = NULL, probes = NULL, silent = TRUE) {
    
    if (is.null(platform)) {
        if (is.null(probes)) {
            platform <- "EPIC" # random guess
        } else {
            platform <- inferPlatformFromProbeIDs(probes, silent = silent)
        }
    }
    stopifnot(platform %in% c(
        "MSA", "EPICv2", "EPIC", "HM27", "HM450", "MM285", "Mammal40"))
    platform
}

## ----message=FALSE, warning=FALSE---------------------------------------------
library(sesameData)
library(GenomicRanges)
sesameDataCache(c("genomeInfo.mm10", "HM450.address"))

## -----------------------------------------------------------------------------
sesameData_getTxnGRanges("mm10")

## -----------------------------------------------------------------------------
sesameData_getTxnGRanges("mm10", merge2gene=TRUE)

## ----message = FALSE, warning = FALSE-----------------------------------------
library(GenomicRanges)

## ----eval=FALSE---------------------------------------------------------------
# ## probes in a region
# sesameData_getProbesByRegion(GRanges('chr5',
#     IRanges(135313937, 135419936)), platform = 'Mammal40')
# ## get chrX probes
# sesameData_getProbesByRegion(chrm = 'chrX', platform = "Mammal40")
# ## get autosomal probes
# sesameData_getProbesByRegion(
#     chrm_to_exclude = c("chrX", "chrY"), platform = "Mammal40")
# ## get DNMT3A probes
# sesameData_getProbesByGene('DNMT3A', "Mammal40", upstream=500)
# ## get DNMT3A promoter probes
# sesameData_getProbesByGene('DNMT3A', "Mammal40", promoter = TRUE)
# ## get all promoter probes
# sesameData_getProbesByGene(NULL, "Mammal40", promoter = TRUE)

## ----eval=FALSE---------------------------------------------------------------
# input_probes <- names(sesameData_getManifestGRanges("Mammal40"))[1:500]
# 
# ## annotate for promoter
# regs <- promoters(sesameData_getTxnGRanges("hg38"))
# sesameData_annoProbes(input_probes, regs, column = "gene_name")
# 
# ## annotate for gene association
# regs <- sesameData_getTxnGRanges("hg38", merge2gene = TRUE)
# sesameData_annoProbes(input_probes, regs, column = "gene_name")
# 
# ## get genes associated with probes
# regs <- sesameData_getTxnGRanges("hg38", merge2gene = TRUE)
# sesameData_annoProbes(input_probes, regs, return_ov_features=TRUE)
# 
# ## get genes associated with probes extending 10kb
# input_probes <- c("cg14620903","cg22464003")
# sesameData_annoProbes(input_probes, regs+10000, column = "gene_name")

## -----------------------------------------------------------------------------
gr <- sesameData_getManifestGRanges("HM450")
length(gr)

## -----------------------------------------------------------------------------
head(sesameDataList())

## -----------------------------------------------------------------------------
sesameDataCache()

## -----------------------------------------------------------------------------
HM27.address <- sesameDataGet('HM27.address')

## -----------------------------------------------------------------------------
sesameDataGet_resetEnv()

## -----------------------------------------------------------------------------
sessionInfo()

#' Extend a GRanges
#'
#' source: https://support.bioconductor.org/p/78652/
#' @param gr a GenomicRanges::GRanges
#' @param upstream distance to expand upstream
#' @param downstream distance to expand downstream
#' @importFrom GenomicRanges trim
#' @importFrom GenomicRanges start
#' @importFrom GenomicRanges end
#' @return a GenomicRanges::GRanges
extend <- function(gr, upstream=0, downstream=0) {
    if (any(strand(gr) == "*")) {
        warning("'*' ranges were treated as '+'") }

    on_plus <- strand(gr) == "+" | strand(gr) == "*"
    ranges(gr) <- IRanges(
        start(gr) - ifelse(on_plus, upstream, downstream),
        end(gr) + ifelse(on_plus, downstream, upstream))
    trim(gr)
}

valid_url <- function(url_in,t=2){
    con <- url(url_in)
    check <- suppressWarnings(try(open.connection(
        con,open="rt",timeout=t),silent=TRUE)[1])
    suppressWarnings(try(close.connection(con),silent=TRUE))
    ifelse(is.null(check),TRUE,FALSE)
}

###############################
######### inst/extdata ########
###############################

## TCGA Ovarian tumor HM27 IDATs
## barcode: TCGA-13-0799-01A-01D-0359-05
## 4207113116_A_Grn.idat
## 4207113116_A_Red.idat

## TCGA Lymphoblastoid cell line control included in Ovarian cancer
## barcode: TCGA-07-0227-20A-01D-0402-05
## 4207113116_B_Grn.idat
## 4207113116_B_Red.idat

## EPIC.sset.LNCaP.Rep1.rds, EPIC.sset.LNCaP.Rep1.chr4.rds
## EPIC.ssets.5normals.rds, EPIC.ssets.5normals.chr4.rds, EPIC.seg.LNCaP.Rep1.rds
## source: GSE86829
library(sesame)
ssets <- readIDATsFromDir('~/tools/sesame/extdata/GSE86829_EPIC_evaluation')
names(ssets) <- gsub('\ ', '.', samplenames[match(substr(names(ssets), 1, 10), samplenames$V1), 'V2'])
EPIC.sset.LNCaP.Rep1 <- ssets[['LNCaP.Rep1']]
EPIC.sset.LNCaP.Rep1.chr4 <- EPIC.sset.LNCaP.Rep1[names(probes[seqnames(probes)=='chr4'])]
EPIC.ssets.5normals <- lapply(ssets[grepl('PrEC', names(ssets)) | grepl('NAF', names(ssets))], function(s) dyeBiasCorr(noob(s)))
EPIC.ssets.5normals.chr4 <- lapply(EPIC.ssets.5normals, function(s) s[names(probes[seqnames(probes)=='chr4'])])
EPIC.seg.LNCaP.Rep1 <- cnSegmentation(EPIC.sset.LNCaP.Rep1.chr4, ssets.normal.chr4)

## HM450.betas.10TCGAnormalPAAD.rds
## HM450.betas.76matchedTCGAchr20.rds
## HM450.sampleinfo.76matchedTCGAchr20.rds
## PAAD    TCGA-FZ-5919-11A-02D-1744-05    6055424077_R04C02
## PAAD    TCGA-FZ-5920-11A-01D-1744-05    6055424077_R05C01
## PAAD    TCGA-FZ-5922-11A-01D-1744-05    6055424077_R02C01
## PAAD    TCGA-FZ-5923-11A-01D-1744-05    6055424077_R03C02
## PAAD    TCGA-FZ-5924-11A-01D-1744-05    6055424077_R04C01
## PAAD    TCGA-FZ-5926-11A-01D-1744-05    6055424077_R06C01
## PAAD    TCGA-H6-8124-11A-01D-2399-05    8795194038_R04C02
## PAAD    TCGA-H6-A45N-11A-12D-A26Q-05    8795194086_R01C02
## PAAD    TCGA-HV-A5A3-11A-11D-A26Q-05    8795194086_R05C02
## PAAD    TCGA-YB-A89D-11A-11D-A368-05    9482801040_R03C01
merged.mapping <- read.table('450k/merged_mapping', sep='\t', stringsAsFactors = F, col.names = c('cancer','barcode','idat'))
merged.mapping.normal <- merged.mapping[substr(merged.mapping$barcode, 14,14) == '1',]
normalPAAD <- merged.mapping.normal[merged.mapping.normal$cancer == 'PAAD',]
betasPAAD <- apply(normalPAAD, 1, function(x) {
    getBetas(dyeBiasCorr(noob(readIDATs(paste0('450k/IDATs/', x[['idat']], '.rda'))[[1]])))
})
colnames(betasPAAD) <- normalPAAD$barcode

## HM450.betas.TCGA-2L-AAQA-01A-21D-A38H-05.rds
## A PAAD tumor
## PAAD    TCGA-2L-AAQA-01A-21D-A38H-05    9806233125_R03C02
HM450.betas.TCGA-2L-AAQA-01A-21D-A38H-05 <- getBetas(dyeBiasCorr(noob(readIDATs('450k/IDATs/9806233125_R03C02.rda'))[[1]]))

## HM450.ssets.10normals.rds, 10 TCGA BLCA adjacent normal samples
## "TCGA-BL-A13J-11A-13D-A10W-05" "TCGA-BT-A20J-11A-11D-A14Z-05"
## "TCGA-BT-A20N-11A-11D-A14Z-05" "TCGA-BT-A20P-11A-11D-A14Z-05"
## "TCGA-BT-A20R-11A-11D-A16P-05" "TCGA-BT-A20U-11A-11D-A14Z-05"
## "TCGA-BT-A20V-11A-11D-A14Z-05" "TCGA-BT-A20W-11A-11D-A14Z-05"
## "TCGA-BT-A20X-11A-12D-A16P-05" "TCGA-BT-A2LA-11A-11D-A18G-05"
ms <- readIDATs(merged.mapping.normal$idat, base.dir='450k/IDATs', raw=T, mc=T)
message('Processing')
ssets <- mclapply(names(dms), function(nm) {
  dm <- dms[[nm]]
  sset <- chipAddressToSignal(dm)
  sset
})
names(ssets) <- idat2barcode[names(dms)]
HM450.ssets.10normal <- ssets[1:10]


###############
### data/ #####
###############

##### cytoband ####
library(biovizBase)
## on R 3.4, not working on R 3.5, should have the package
## import biovizBase when this is fixed
cytoband.hg19 <- as.data.frame(getIdeogram('hg19', cytoband = TRUE))
cytoband.hg38 <- as.data.frame(getIdeogram('hg38', cytoband = TRUE))

library(rtracklayer)
sess <- browserSession('UCSC')
genome(sess) <- 'hg19'
cytoBand <- getTable(ucscTableQuery(sess, table='cytoBand'))
saveRDS(cytoBand, 'sesame/latest/cytoBand.hg19.rds')

genome(sess) <- 'hg38'
cytoBand <- getTable(ucscTableQuery(sess, table='cytoBand'))
saveRDS(cytoBand, 'sesame/latest/cytoBand.hg38.rds')

##### ordering #####
a <- readRDS(gzcon(url('http://zwdzwd.io/InfiniumAnnotation/current/EPIC/EPIC.hg19.manifest.rds')))
EPIC.ordering <- data.frame(Probe_ID=names(a), M=a$address_B, U=a$address_A, DESIGN=a$designType, COLOR_CHANNEL=a$channel, col=as.factor(ifelse(a$designType=='I', substr(a$channel,1,1), NA)))

a <- readRDS(gzcon(url('http://zwdzwd.io/InfiniumAnnotation/current/hm450/hm450.hg19.manifest.rds')))
HM450.ordering <- data.frame(Probe_ID=names(a), M=a$address_B, U=a$address_A, DESIGN=a$designType, COLOR_CHANNEL=a$channel, col=as.factor(ifelse(a$designType=='I', substr(a$channel,1,1), NA)))

a <- readRDS(gzcon(url('http://zwdzwd.io/InfiniumAnnotation/current/hm27/hm27.hg19.manifest.rds')))
HM27.ordering <- data.frame(Probe_ID=names(a), M=a$address_B, U=a$address_A, DESIGN=a$designType, COLOR_CHANNEL=a$channel, col=as.factor(ifelse(a$designType=='I', substr(a$channel,1,1), NA)))

##### controls ######
manifest <- read.csv('EPIC.official.manifest/MethylationEPIC_v-1-0_B4_noheader.csv', header=T, stringsAsFactors=F, row.names=1)
ct <- which(rownames(manifest) == '[Controls]')
manifest.controls <- manifest[(ct+1):nrow(manifest),]
EPIC.controls <- with(manifest.controls, data.frame(Address=rownames(manifest.controls), Type=Name, Color_Channel=AddressA_ID, Name=AlleleA_ProbeSeq))

##### non-pseudoautosomal chrY probes ######
load('/secondary/projects/shen/projects/2016_07_01_dyebias/allnormal.MPU.chrY.rda')
load('/secondary/projects/shen/projects/2016_07_01_dyebias/allnormal.pvaloob.chrY.rda')
samplemean <-  apply(allnormal.MPU.chrY, 2, mean)
femalesamples <- names(samplemean[samplemean<3000])
malesamples <- names(samplemean)[!(names(samplemean) %in% femalesamples)]
female.chrY.numdetect <- apply(allnormal.pvaloob.chrY[,femalesamples], 1, function(x) {sum(x<0.05)})
head(sort(female.chrY.numdetect, decreasing=T), n=100)
hm450.female.clean.chrY.probes <- names(female.chrY.numdetect)[female.chrY.numdetect <= 5]
head(hm450.female.clean.chrY.probes)
length(hm450.female.clean.chrY.probes) # 337 probes
save(hm450.female.clean.chrY.probes, file='~/tools/sesame/sesame/data/hm450.female.clean.chrY.probes.rda')
EPIC.female.clean.chrY.probes <- female.clean.chrY.probes[female.clean.chrY.probes %in% names(EPIC.manifest.hg38)]
length(EPIC.female.clean.chrY.probes) # 314 probes
save(EPIC.female.clean.chrY.probes, file='~/tools/sesame/sesame/data/EPIC.female.clean.chrY.probes.rda')

##### X-linked probes #####
load('/secondary/projects/shen/projects/2016_07_01_dyebias/allnormal.betas.chrX.rda')
head(allnormal.betas.chrX[,femalesamples])
head(allnormal.betas.chrX[,malesamples])
allnormal.betas.chrX <- allnormal.betas.chrX[grep('cg',rownames(allnormal.betas.chrX)),]
female.chrX.mediumCnt <- apply(allnormal.betas.chrX[,femalesamples], 1, function(x) sum(x>0.3 & x<0.7, na.rm = T))
hm450.female.xlinked.chrX.probes <- names(female.chrX.mediumCnt[female.chrX.mediumCnt > 300])
length(hm450.female.xlinked.chrX.probes) # 3797 probes
save(hm450.female.xlinked.chrX.probes, file='~/tools/sesame/sesame/data/hm450.female.xlinked.chrX.probes.rda')
EPIC.female.xlinked.chrX.probes <- female.xlinked.chrX.probes[female.xlinked.chrX.probes %in% names(EPIC.manifest.hg38)]
length(EPIC.female.xlinked.chrX.probes) # 3433 probes
save(EPIC.female.xlinked.chrX.probes, file='~/tools/sesame/sesame/data/EPIC.female.xlinked.chrX.probes.rda')

###### probe2chr #######
EPIC.hg19.probe2chr <- setNames(with(manifest.ordering, as.factor(ifelse(CHR=="", as.character(rschr), as.character(paste0('chr',CHR))))),rownames(manifest.ordering))
more.missing <- which(is.na(EPIC.hg19.probe2chr))
load('data/hm450.hg19.probe2chr.rda')
EPIC.hg19.probe2chr[more.missing] <- hm450.hg19.probe2chr[names(more.missing)]
any(is.na(EPIC.hg19.probe2chr)) # no NA
save(EPIC.hg19.probe2chr, file='data/EPIC.hg19.probe2chr.rda')

####### probe mapping #######
load('GR.InfiniumMethylation/20160711//EPIC/EPIC.manifest.hg38.rda')
mani <- EPIC.manifest.hg38
mani <- mani[!mani$MASK.mapping]
probedf <- as.data.frame(mani)
probedf$seqnames <- as.character(probedf$seqnames)
probedf <- probedf[probedf$seqnames != '*',]
EPIC.mapped.probes.hg38 <- GRanges(as.character(probedf$seqnames), IRanges(probedf$start, probedf$start), seqinfo=hg38.chrominfo$seqinfo)
names(EPIC.mapped.probes.hg38) <- rownames(probedf)

load('GR.InfiniumMethylation/20160711//EPIC/EPIC.manifest.rda')
mani <- EPIC.manifest
mani <- mani[!mani$MASK.mapping]
probedf <- as.data.frame(mani)
probedf$seqnames <- as.character(probedf$seqnames)
probedf <- probedf[probedf$seqnames != '*',]
EPIC.mapped.probes.hg19 <- GRanges(as.character(probedf$seqnames), IRanges(probedf$start, probedf$start), seqinfo=hg19.chrominfo$seqinfo)
names(EPIC.mapped.probes.hg19) <- rownames(probedf)

load('GR.InfiniumMethylation/20160711//hm450/hm450.manifest.hg38.rda')
mani <- hm450.manifest.hg38
mani <- mani[!mani$MASK.mapping]
probedf <- as.data.frame(mani)
probedf$seqnames <- as.character(probedf$seqnames)
probedf <- probedf[probedf$seqnames != '*',]
hm450.mapped.probes.hg38 <- GRanges(as.character(probedf$seqnames), IRanges(probedf$start, probedf$start), seqinfo=hg38.chrominfo$seqinfo)
names(hm450.mapped.probes.hg38) <- rownames(probedf)

load('GR.InfiniumMethylation/20160711//hm450/hm450.manifest.rda')
mani <- hm450.manifest
mani <- mani[!mani$MASK.mapping]
probedf <- as.data.frame(mani)
probedf$seqnames <- as.character(probedf$seqnames)
probedf <- probedf[probedf$seqnames != '*',]
hm450.mapped.probes.hg19 <- GRanges(as.character(probedf$seqnames), IRanges(probedf$start, probedf$start), seqinfo=hg19.chrominfo$seqinfo)
names(hm450.mapped.probes.hg19) <- rownames(probedf)

####### mask #######
EPIC.manifest <- readRDS(gzcon(url('http://zwdzwd.io/InfiniumAnnotation/current/EPIC/EPIC.hg19.manifest.rds')))
EPIC.mask <- names(EPIC.manifest[EPIC.manifest$MASK.general])

hm450.manifest <- readRDS(gzcon(url('http://zwdzwd.io/InfiniumAnnotation/current/hm450/hm450.hg19.manifest.rds')))
hm450.mask <- names(hm450.manifest[hm450.manifest$MASK.general])

hm27.manifest <- readRDS(gzcon(url('http://zwdzwd.io/InfiniumAnnotation/current/hm27/hm27.hg19.manifest.rds')))
hm27.mask <- names(hm27.manifest[hm27.manifest$MASK.general])

######## C/T extension probes ########
load('GR.InfiniumMethylation/latest//EPIC/EPIC.manifest.hg38.rda')
EPIC.typeI.extC <- names(EPIC.manifest.hg38[(!EPIC.manifest.hg38$MASK.general) & EPIC.manifest.hg38$designType=='I' & EPIC.manifest.hg38$nextBaseRef=='C'])
EPIC.typeI.extT <- names(EPIC.manifest.hg38[(!EPIC.manifest.hg38$MASK.general) & EPIC.manifest.hg38$designType=='I' & EPIC.manifest.hg38$nextBaseRef=='T'])
load('GR.InfiniumMethylation/latest//hm450/hm450.manifest.hg38.rda')
hm450.typeI.extC <- names(hm450.manifest.hg38[(!hm450.manifest.hg38$MASK.general) & hm450.manifest.hg38$designType=='I' & hm450.manifest.hg38$nextBaseRef=='C'])
hm450.typeI.extT <- names(hm450.manifest.hg38[(!hm450.manifest.hg38$MASK.general) & hm450.manifest.hg38$designType=='I' & hm450.manifest.hg38$nextBaseRef=='T'])

######## ethnicity, probes and models #######
load('GR.InfiniumMethylation/20160711//EPIC/EPIC.manifest.rda')
load('GR.InfiniumMethylation/20160711//hm450//hm450.manifest.rda')
rsprobes <- intersect(names(EPIC.manifest)[grep('rs', names(EPIC.manifest))], names(hm450.manifest)[grep('rs', names(hm450.manifest))])
rsprobes <- sort(rsprobes)
ccsprobes <- colnames(SNPswitched.maf001)
ccsprobes <- sort(ccsprobes)
samples <- intersect(rownames(SNPswitched.maf001), colnames(rs.betas))
samples <- samples[!is.na(samplerace[samples])]

df.train <- cbind(t(rs.betas[rsprobes,samples]), SNPswitched.maf001[samples, ccsprobes])
df.train[is.na(df.train)] <- 0.5
library(randomForest)
fit <- randomForest(x=df.train, y=as.factor(samplerace[rownames(df.train)]), importance=TRUE, ntree=200)
load('450k/signals.dyebias/3999492009_R01C02.rda')
load('450k/signals/3999492009_R01C02.rda')
sset[rsprobes]
b <- sset[ccsprobes]$toBetaTypeIbySum(na.mask=FALSE)

a <- getBetas(sset[rsprobes], quality.mask = F, nondetection.mask=F)
b <- getBetasTypeIbySumAlleles(sset[ccsprobes], quality.mask = F, nondetection.mask = F)
ab <- c(a,b)
predict(fit, ab)

ethnicity.ccs.probes <- ccsprobes
ethnicity.rs.probes <- rsprobes
save(ethnicity.ccs.probes, file='~/tools/sesame/sesame/data/ethnicity.ccs.probes.rda')
save(ethnicity.rs.probes, file='~/tools/sesame/sesame/data/ethnicity.rs.probes.rda')
ethnicity.model <- fit
save(ethnicity.model, file='~/tools/sesame/sesame/data/ethnicity.model.rda')

######## chrominfo #########
library(rtracklayer)
sess <- browserSession('UCSC')
## set/get genome build
genome(sess) <- 'hg19'
tableQuery <- ucscTableQuery(sess)
## list all table names:
## tableNames(tableQuery)
seqnames <- paste0('chr',c(1:22,'X','Y','M'))
chromInfo <- getTable(ucscTableQuery(sess, table='chromInfo'))
chromInfo <- chromInfo[chromInfo$chrom %in% seqnames,]
head(chromInfo)
seqlengths <- chromInfo$size[match(seqnames, chromInfo$chrom)]
seqinfo <- Seqinfo(seqnames, seqlengths)
gap <- getTable(ucscTableQuery(sess, table='gap'))
gap <- gap[gap$chrom %in% seqnames,]
cytoBand <- getTable(ucscTableQuery(sess, table='cytoBand'))
cytoBand <- cytoBand[cytoBand$chrom %in% seqnames,]
gap <- sort(GRanges(as.vector(gap$chrom), IRanges(gap$chromStart+1, gap$chromEnd), seqinfo=seqinfo))
hg19.chrominfo <- list(gap=gap, seqinfo=seqinfo)
save(hg19.chrominfo, file='~/tools/sesame/sesame/data/hg19.chrominfo.rda', compress='xz')

genome(sess) <- 'hg38'
tableQuery <- ucscTableQuery(sess)
seqnames <- paste0('chr',c(1:22,'X','Y','M'))
chromInfo <- getTable(ucscTableQuery(sess, table='chromInfo'))
chromInfo <- chromInfo[chromInfo$chrom %in% seqnames,]
head(chromInfo)
seqlengths <- chromInfo$size[match(seqnames, chromInfo$chrom)]
seqinfo <- Seqinfo(seqnames, seqlengths)
gap <- getTable(ucscTableQuery(sess, table='gap'))
gap <- gap[gap$chrom %in% seqnames,]
head(gap)
cytoBand <- getTable(ucscTableQuery(sess, table='cytoBand'))
cytoBand <- cytoBand[cytoBand$chrom %in% seqnames,]
head(cytoBand)
gap <- sort(GRanges(as.vector(gap$chrom), IRanges(gap$chromStart+1, gap$chromEnd), seqinfo=seqinfo))
hg38.chrominfo <- list(gap=gap, seqinfo=seqinfo)
save(hg38.chrominfo, file='~/tools/sesame/sesame/data/hg38.chrominfo.rda', compress='xz')

######### transcripts ##########
library(GenomicRanges)
refgene <- read.table('/references/hg19/annotations/UCSC_refGene_hg19_20161101.tsv', stringsAsFactors = F, check.names = F, header=T, comment.char = '')
colnames(refgene)
txns <- do.call(GRangesList, apply(refgene, 1, function(row) {
  g <- GRanges(row[['chrom']],
        ranges=IRanges(sapply(unlist(strsplit(row[['exonStarts']],',')), as.integer),
          sapply(unlist(strsplit(row[['exonEnds']],',')), as.integer)),
        strand=row[['strand']])
  mcols(g)$cdsStart <- row[['cdsStart']]
  mcols(g)$cdsEnd <- row[['cdsEnd']]
  sort(g)
}))

txn2gene <- split(refgene$name2, refgene$name)
gene2txn <- split(refgene$name, refgene$name2)
names(txns) <- refgene$name

saveRDS(txns, file='sesame/latest/UCSC.refGene.txns.hg19.rds')
saveRDS(txn2gene, file='sesame/latest/UCSC.refGene.txn2gene.hg19.rds')
saveRDS(gene2txn, file='sesame/latest/UCSC.refGene.gene2txn.hg19.rds')

library(GenomicRanges)
refgene <- read.table('/references/hg38/annotations/UCSC_refGene_hg38_20161101.tsv', stringsAsFactors = F, check.names = F, header=T, comment.char = '')
colnames(refgene)
txns <- do.call(GRangesList, apply(refgene, 1, function(row) {
  g <- GRanges(row[['chrom']],
        ranges=IRanges(sapply(unlist(strsplit(row[['exonStarts']],',')), as.integer),
          sapply(unlist(strsplit(row[['exonEnds']],',')), as.integer)),
        strand=row[['strand']])
  mcols(g)$cdsStart <- as.integer(row[['cdsStart']])
  mcols(g)$cdsEnd <- row[['cdsEnd']])
  sort(g)
}))

txn2gene <- split(refgene$name2, refgene$name)
gene2txn <- split(refgene$name, refgene$name2)
names(txns) <- refgene$name
saveRDS(txns, file='sesame/latest/UCSC.refGene.txns.hg38.rds')
saveRDS(txn2gene, file='sesame/latest/UCSC.refGene.txn2gene.hg38.rds')
saveRDS(gene2txn, file='sesame/latest/UCSC.refGene.gene2txn.hg38.rds')

######### age #########
a <- read.csv('Rtutorial/AdditionalFile3.csv',header=T)
saveRDS(a, file='/secondary/projects/shen/projects/2016_12_06_sesame_home/Horvath353.rds')

######### cell type reference ########
load('GR.InfiniumMethylation/20160711/hm450/hm450.manifest.hg38.rda')
load('GR.InfiniumMethylation/20160711/EPIC/EPIC.manifest.hg38.rda')

discretize <- function(m) {
  md <- as.numeric(cut(m, c(-1,0.3,0.7,1)))
  md <- matrix(md, ncol=ncol(m))
  dimnames(md) <- dimnames(m)
  md
}

construct.reference <- function(m) {
  md <- discretize(m)
  nsamples <- ncol(m)
  cnt <- apply(md, 1, function(x) sum(x==3))
  g <- setNames(rep(NA, nrow(m)), rownames(m))
  g[names(na.omit(cnt[cnt==ncol(m)]))] <- 1
  cnt <- apply(md, 1, function(x) sum(x==1))
  g[names(na.omit(cnt[cnt==ncol(m)]))] <- 0
  g
}

liftEPIC <- function(g) {
  aa <- match(names(g), EPIC.probes)
  ge <- EPIC.sig.na
  ge[na.omit(aa)] <- g[!is.na(aa)]
  ge
}

## relaxed construction, tune uniform.frac to set
## the minimum uniformity in setting the state of a probe
construct.reference.relaxed <- function(m, uniform.frac=0.8) {
  md <- discretize(m)
  nsamples <- ncol(m)
  cnt <- apply(md, 1, function(x) sum(x==3))
  g <- setNames(rep(NA, nrow(m)), rownames(m))
  g[names(na.omit(cnt[cnt > uniform.frac*ncol(m)]))] <- 1
  cnt <- apply(md, 1, function(x) sum(x==1))
  g[names(na.omit(cnt[cnt > uniform.frac*ncol(m)]))] <- 0
  g
}

construction.pipeline <- function(betas, cls, clsname) {
  b <- betas[,cls]
  attr(b, 'tissue') <- clsname
  saveRDS(b, file=sprintf('curation_betas/hm450.%s.rds', clsname))
  saveRDS(liftEPICb(b), file=sprintf('curation_betas/EPIC.%s.rds', clsname))
  g <- construct.reference(b)
  message(sum(is.na(g)))
  attr(g, 'tissue') <- 'granulocytes'
  saveRDS(g, file=sprintf('curation/hm450.%s.rds', clsname))
  saveRDS(liftEPIC(g), file=sprintf('curation/EPIC.%s.rds', clsname))
}

construction.pipeline(betas, c('Granulocytes_5684819001_R01C02', 'Granulocytes_5684819001_R02C02', 'Granulocytes_5684819001_R03C02'), 'granulocytes')
construction.pipeline(betas, c('CD4+_T-cells_5727920027_R04C02', 'CD4+_T-cells_5727920027_R05C02', 'CD4+_T-cells_5727920027_R06C02'), 'CD4T')
construction.pipeline(betas, c('CD8+_T-cells_5727920033_R01C01', 'CD8+_T-cells_5727920033_R02C01', 'CD8+_T-cells_5727920033_R03C01'), 'CD8T')
construction.pipeline(betas, c('CD14+_Monocytes_5727920033_R01C02', 'CD14+_Monocytes_5727920033_R02C02', 'CD14+_Monocytes_5727920033_R03C02'), 'CD14Monocytes')
construction.pipeline(betas, c('CD56+_NK-cells_5727920033_R04C02', 'CD56+_NK-cells_5727920033_R05C02', 'CD56+_NK-cells_5727920033_R06C02'), 'CD56NK')
construction.pipeline(betas, c('CD19+_B-cells_5727920033_R04C01', 'CD19+_B-cells_5727920033_R05C01', 'CD19+_B-cells_5727920033_R06C01'), 'CD19B')
construction.pipeline(betas, c('Eosinophils_5727920038_R01C02', 'Eosinophils_5727920038_R02C02', 'Eosinophils_5727920038_R03C02', 'Eosinophils_5727920038_R04C02', 'Eosinophils_5727920038_R05C02', 'Eosinophils_5727920038_R06C02'), 'eosinophil')
construction.pipeline(betas, c('Neutrophils_5727920038_R01C01', 'Neutrophils_5727920038_R02C01', 'Neutrophils_5727920038_R03C01', 'Neutrophils_5727920038_R04C01', 'Neutrophils_5727920038_R05C01', 'Neutrophils_5727920038_R06C01'), 'neutrophil')
construction.pipeline(betas, c('Hair_PT1', 'Hair_PT2', 'Hair_PT3', 'Hair_PT4', 'Hair_PT5'), 'hair') # 168536
construction.pipeline(betas, c('Muscle_IT9', 'Muscle_IT13', 'Muscle_IT10', 'Muscle_IT15', 'Muscle_IT16', 'Muscle_IT17'), 'muscle') # 193000
construction.pipeline(betas, c('SC_Fat_IT9', 'SC_Fat_IT13', 'SC_Fat_IT10', 'SC_Fat_IT15', 'SC_Fat_IT16', 'SC_Fat_IT17'), 'scFat') # 201708
construction.pipeline(betas, c('Omentum_IT9', 'Omentum_IT13', 'Omentum_IT10', 'Omentum_IT15', 'Omentum_IT16', 'Omentum_IT17'), 'omentum') # 207868
construction.pipeline(betas, c('Saliva_PT1', 'Saliva_PT2', 'Saliva_PT3', 'Saliva_PT4', 'Saliva_PT5'), 'saliva') # 161780
construction.pipeline(betas, c('Liver_IT9', 'Liver_IT13', 'Liver_IT10', 'Liver_IT15', 'Liver_IT17'), 'liver') # 192919
construction.pipeline(betas, c('Spleen_IT9', 'Spleen_IT10', 'Spleen_IT16'), 'spleen') # 182984
construction.pipeline(betas, c('Buccal_PT1', 'Buccal_PT2', 'Buccal_PT3', 'Buccal_PT4', 'Buccal_PT5'), 'buccal') # 185727

load('dataset/skin.rda')
colnames(betas)
b <- betas
attr(b, 'tissue') <- 'skin'
saveRDS(b, file='curation_betas/hm450.skin.rds')
saveRDS(liftEPICb(b), file='curation_betas/EPIC.skin.rds')
g <- construct.reference.relaxed(b)
sum(is.na(g)) # 204920 a bit high, need to relax
attr(g, 'tissue') <- 'skin'
saveRDS(g, file='curation/hm450.skin.rds')
saveRDS(liftEPIC(g), file='curation/EPIC.skin.rds')

library(readxl)
metadata <- as.data.frame(read_excel('~/Dropbox/sesame/sesameData/inst/extdata/metadata0.xlsx', sheet='metadata0'))
colnames(metadata) <- c(
    'Title', 'Description', 'BiocVersion', 'Genome', 'SourceType', 'SourceUrl',
    'SourceVersion', 'Species', 'TaxonomyId', 'Coordinate_1_based', 'DataProvider',
    'Maintainer', 'RDataClass', 'DispatchClass', 'RDataPath', 'Tags', 'Notes')

metadata$BiocVersion <- as.numeric(metadata$BiocVersion)
metadata$TaxonomyId <- as.numeric(metadata$TaxonomyId)
metadata$Coordinate_1_based <- as.logical(metadata$Coordinate_1_based)

write.csv(metadata, file='~/tools/sesame/sesameData/inst/extdata/metadata.csv', row.names = FALSE)

#' Get probes by genomic region
#'
#' The function takes a genomic coordinate and output the a vector of probes
#' on the specified platform that falls in the given genomic region.
#'
#' @param regs GRanges
#' @param chrm chromosome, when given regs are ignored
#' @param chrm_to_exclude chromosome to exclude.
#' @param beg begin, 1 if omitted
#' @param end end, chromosome end if omitted
#' @param platform EPICv2, EPIC, HM450, ...
#' @param genome hg38, mm10, ... will infer if not given.
#' For additional mapping, download the GRanges object from
#' http://zwdzwd.github.io/InfiniumAnnotation
#' and provide the following argument
#' ..., genome = sesameAnno_buildManifestGRanges("downloaded_file"),...
#' to this function.
#' @return GRanges of selected probes
#' @importMethodsFrom IRanges subsetByOverlaps
#' @examples
#'
#' ## download needed data
#' sesameDataCache(c("Mammal40.address", "genomeInfo.hg38"))
#' 
#' ## get probes in a region
#' library(GenomicRanges)
#' probes = sesameData_getProbesByRegion(
#'     GRanges('chr5', IRanges(135313937, 135419936)), platform = 'Mammal40')
#'
#' ## get all probes on chromosome 5
#' probes = sesameData_getProbesByRegion(chrm = "chr5", platform = "Mammal40")
#'
#' ## get all probes on chromosome X
#' probes = sesameData_getProbesByRegion(chrm = 'chrX', platform = "Mammal40")
#'
#' ## get all probes on both chromosome X and Y
#' probes = sesameData_getProbesByRegion(
#'     chrm = c('chrX', 'chrY'), platform = "Mammal40")
#'
#' ## get all autosomal probes
#' probes = sesameData_getProbesByRegion(
#'     chrm_to_exclude = c("chrX", "chrY"), platform = "Mammal40")
#' 
#' @export
sesameData_getProbesByRegion <- function(
    regs, chrm = NULL, beg = 1, end = -1,
    platform = NULL, chrm_to_exclude = NULL, genome = NULL) {

    platform <- sesameData_check_platform(platform)
    genome <- sesameData_check_genome(genome, platform)
    probes <- sesameData_getManifestGRanges(platform, genome=genome)

    if (!is.null(chrm) || !is.null(chrm_to_exclude)) {
        if (beg == 1 && end < 0) {
            if (is.null(chrm_to_exclude)) {
                return(probes[as.character(
                    GenomicRanges::seqnames(probes)) %in% chrm])
            } else {
                return(probes[!(as.character(
                    GenomicRanges::seqnames(probes)) %in% chrm_to_exclude)])
            }
        }
        if (end < 0) {
            seqLength <- sesameData_getGenomeInfo(genome)$seqLength
            stopifnot(chrm %in% names(seqLength))
            end <- seqLength[chrm]
        }
        message(sprintf('Retrieve probes from %s:%d-%d.\n', chrm, beg, end))
        regs <- GenomicRanges::GRanges(chrm, IRanges::IRanges(beg, end))
    }
    subsetByOverlaps(probes, regs)
}

#' Get Probes by Genes or Gene Promoters
#'
#' Get probes mapped to a gene. All transcripts for the gene are considered.
#' The function takes a gene name as appears in UCSC RefGene database. The
#' platform and reference genome build can be changed with `platform` and
#' `genome` options. The function returns a vector of probes that falls
#' into the given gene.
#'
#' @param gene_name gene name, if NULL return all genes
#' @param platform EPIC or HM450
#' @param promoter if TRUE, use TSS instead of the whole gene
#' @param upstream number of bases to expand upstream of target gene
#' @param downstream number of bases to expand downstream of target gene
#' @param genome hg38 or hg19
#' @return GRanges containing probes that fall into the given gene
#' @examples
#'
#' ## download needed data
#' sesameDataCache(c("Mammal40.address", "genomeInfo.hg38"))
#' 
#' ## get all probes overlapping with DNMT3A
#' probes <- sesameData_getProbesByGene(
#'     'DNMT3A', "Mammal40", upstream=500, downstream=500)
#'
#' ## get the promoter-associated probes
#' probes <- sesameData_getProbesByGene('DNMT3A', "Mammal40", promoter = TRUE)
#' 
#' @export
sesameData_getProbesByGene <- function(
    gene_name = NULL, platform = NULL, promoter = FALSE,
    upstream = 1500, downstream = 1500, genome = NULL) {

    platform <- sesameData_check_platform(platform)
    genome <- sesameData_check_genome(genome, platform)
    
    txns <- sesameData_getTxnGRanges(genome)
    if (!is.null(gene_name)) {
        txns <- txns[txns$gene_name == gene_name]
    }
    stopifnot(length(txns) > 0)

    if (promoter) {
        reg <- promoters(
            txns, upstream = upstream, downstream = downstream)
    } else {
        up <- ifelse(as.vector(GenomicRanges::strand(
            txns)) == '-', downstream, upstream)
        dw <- ifelse(as.vector(GenomicRanges::strand(
            txns)) == '-', upstream, downstream)
        reg <- GRanges(
            as.vector(GenomicRanges::seqnames(txns)),
            IRanges(
                GenomicRanges::start(txns) - up,
                GenomicRanges::end(txns) + dw))
    }
    
    sesameData_getProbesByRegion(reg, platform = platform, genome = genome)
}

#' get Infinium manifest GRanges
#'
#' Note that some unaligned probes are not included.
#' For full manifest, please visit
#' \url{http://zwdzwd.github.io/InfiniumAnnotation}
#'
#' @param platform Mammal40, MM285, EPIC, and HM450
#' @param genome hg38, mm10, ... will infer if not given.
#' For additional mapping, download the GRanges object from
#' http://zwdzwd.github.io/InfiniumAnnotation
#' and provide the following argument
#' ..., genome = sesameAnno_buildManifestGRanges("downloaded_file"),...
#' to this function.
#' @return GRanges
#' @examples
#' sesameDataCache("Mammal40.address")
#' res <- sesameData_getManifestGRanges("Mammal40")
#' @export
sesameData_getManifestGRanges <- function(
    platform, genome = NULL) {

    if ("GenomicRanges" %in% class(genome)) { # if is a GRanges object already
        return(genome)
    }

    platform <- sesameData_check_platform(platform)
    genome <- sesameData_check_genome(genome, platform)
    ## only one genome is supported natively.
    addr <- sesameDataGet(sprintf("%s.address", platform))
    if (genome %in% names(addr)) {
        return(addr[[genome]])
    } else {
        return(NULL)
    }
}

#' Master data frame for all object to cache
#'
#' This is an internal object which will be updated on every new release
#' library(ExperimentHub)
#' eh <- query(ExperimentHub(localHub=FALSE), c("sesameData", "v1.13.1"))
#' data.frame(name=eh$title, eh=names(eh))
#'
#' Cache location is default to
#' /Users/zhouw3/Library/Caches/org.R-project.R/R/ExperimentHub/
#'
#' @format A data frame with 22 columns:
#' \describe{
#'   \item{Comments}{Additional comments}
#'   \item{EHID}{ExperimentHub ID}
#'   \item{VERSION}{sesameData version}
#'   \item{IN_USE}{Logical indicating if the resource is in use}
#'   \item{Title}{Title of the data resource}
#'   \item{Description}{Description of the data resource}
#'   \item{BiocVersion}{Bioconductor version}
#'   \item{Genome}{Genome build (e.g., hg38, mm10)}
#'   \item{SourceType}{Source file type}
#'   \item{SourceUrl}{URL to source}
#'   \item{SourceVersion}{Version of source data}
#'   \item{Species}{Species name}
#'   \item{TaxonomyId}{NCBI Taxonomy ID}
#'   \item{Coordinate_1_based}{Logical indicating if coordinates are 1-based}
#'   \item{DataProvider}{Data provider name}
#'   \item{Maintainer}{Maintainer contact information}
#'   \item{RDataClass}{R data class}
#'   \item{DispatchClass}{Dispatch class for ExperimentHub}
#'   \item{RDataPath}{Path to RData file}
#'   \item{Location_Prefix}{URL prefix for data location}
#'   \item{Tags}{Tags for categorization}
#'   \item{Notes}{Additional notes}
#' }
#' @name df_master
#' @docType data
#' @return master sheet of sesameData objects
NULL

cacheEnv <- new.env()
alt_base <- "https://zhouserver.research.chop.edu"

#' Check whether the title exists in cacheEnv
#'
#' @param title the title to check
#' @return the data associated with the title or NULL if title doesn't exist
sesameDataGet_checkEnv <- function(title) {
    if (exists(title, envir = cacheEnv, inherits = FALSE)) {
        return(get(title, envir = cacheEnv, inherits = FALSE))
    } else {
        return(NULL)
    }
}

sesameDataGet_assignEnv <- function(title, data) {
    assign(title, data, envir = cacheEnv)
}

#' Empty cache environment to free memory
#'
#' When this function is called sesameDataGet will
#' retrieve all data from disk again instead of using the in-memory
#' cache, i.e., sesameData:::cacheEnv.
#'
#' Note this is different from sesameDataClearHub which empties the
#' ExperimentHub on disk.
#'
#' @return gc() output
#' @examples
#' sesameDataGet_resetEnv()
#' @export
sesameDataGet_resetEnv <- function() {
    rm(list=ls(envir=cacheEnv), envir=cacheEnv)
    gc()
}

## fall back to zenodo in case EHID is not assigned yet.
.sesameDataGet_fallback <- function(title, u1) {
    if (valid_url(u1)) {
        sesameDataGet_assignEnv(title, get(load(url(u1))))
        TRUE
    } else {
        warning(sprintf("Resource %s cannot be retrieved.", title))
        FALSE
    }
    TRUE
}

.sesameDataGet <- function(title) {

    idx <- match(title, df_master$Title)
    eh_id <- df_master$EHID[idx]
    stopifnot(length(eh_id) == 1)
    pfx <- df_master$Location_Prefix[idx]
    rdata <- df_master$RDataPath[idx]
    if (is.na(eh_id) && !is.na(pfx)) { # no EHID and exists zenodo link
        ## cache uses title, not ehid
        if (exists(title, envir=cacheEnv, inherits=FALSE)) {
            return(get(title, envir=cacheEnv, inherits=FALSE))
        }
        
        if (.sesameDataGet_fallback(title, paste0(pfx, "/", rdata))) {
            return(get(title, envir=cacheEnv, inherits=FALSE))
        } else {
            stopAndCache(title)
        }
    } else {
        if (exists(eh_id, envir=cacheEnv, inherits=FALSE)) {
            return(get(eh_id, envir=cacheEnv, inherits=FALSE))
        }
        if (!file.exists(getExperimentHubOption("CACHE"))) {
            stopAndCache(title) }
        tryCatch({
            eh <- query(ExperimentHub(localHub=TRUE), 'sesameData')
        }, error = function(cond) { stopAndCache(title); })
        if (!(eh_id %in% names(eh))) {
            stopAndCache(title); }
        sesameDataGet_assignEnv(eh_id, eh[[eh_id]])
        return(get(eh_id, envir=cacheEnv, inherits=FALSE))
    }
}

#' Get SeSAMe data
#'
#' @param title title of the data
#' @param verbose whether to output ExperimentHub message
#' @return data object
#' @import ExperimentHub
#' @import AnnotationHub
#' @importFrom stringr str_replace
#' @examples
#'
#' sesameDataCache("EPIC.1.SigDF")
#' EPIC.1.SigDF <- sesameDataGet('EPIC.1.SigDF')
#' @export
sesameDataGet <- function(title, verbose = FALSE) {

    title <- str_replace(title, "MMB", "MM285") # fix potential code discrepancy
    
    if (verbose) {
        .sesameDataGet(title)
    } else {
        suppressMessages(
            log <- capture.output(obj <- .sesameDataGet(title)))
        obj
    }
}

#' List all SeSAMe data
#'
#' @param filter keyword to filter title, optional
#' @param full whether to display all columns
#' @return all titles from SeSAMe Data
#' @examples
#' sesameDataList("KYCG")
#' @export
sesameDataList <- function(filter = NULL, full = FALSE) {
    df <- df_master
    if (!full) {
        df <- df[,c("EHID","Title")]
    }
    if (!is.null(filter)) {
        df <- df[grep(filter, df$Title),]
    }
    df
}

#' Whether sesameData has
#'
#' @param data_titles data titles to check
#' @return a boolean vector the same length as data_titles
#' @examples
#' sesameDataHas(c("EPIC.address","EPIC.address.Nonexist"))
#' @export
sesameDataHas <- function(data_titles) {
    data_titles %in% sesameDataList()$Title
}

#' Find genome assembly version(s) supported for a platform
#'
#' @param genome mm10, hg38, ..., or NULL
#' @param platform HM27, HM450, EPIC, EPICv2, MSA, ...
#' @return genome as string
#' @examples
#' sesameData_check_genome(NULL, "Mammal40")
#' @export
sesameData_check_genome <- function(genome, platform) {
    if ("GRanges" %in% class(genome)) { return(genome); }
    platform <- sesameData_check_platform(platform)
    supported_genomes <- c("hg19", "hg38", "mm10")
    default_genome <- c(
        HM27 = "hg38", HM450 = "hg38",
        EPIC = "hg38", EPICv2 = "hg38", MSA = "hg38",
        Mammal40 = "hg38", MM285 = "mm10")
    if (is.null(genome)) { genome <- default_genome[platform] }
    stopifnot(!is.null(genome))
    stopifnot(genome %in% supported_genomes)
    genome
}

#' Get genome info files
#' 
#' @param genome hg38, mm10, or GRanges with a metadata(genome)[["genome"]]
#' @return a list of genome info files
#' @examples
#' sesameDataCache("genomeInfo.hg38")
#' res <- sesameData_getGenomeInfo("hg38")
#' @export
sesameData_getGenomeInfo <- function(genome) {
    if ("GenomicRanges" %in% class(genome)) {
        genome <- attr(genome, "genome")
    }
    stopifnot(is.character(genome))
    key <- paste0('genomeInfo.', genome)
    if (!sesameDataHas(key)) {
        stop(sprintf("%s genome info is not found in Bioconductor.
Please go to http://zwdzwd.github.io/InfiniumAnnotation
for additional genome info files.
", genome))
    }
    sesameDataGet(key)
}

read_GENCODE_gtf <- function(x) {
    ## https://www.gencodegenes.org/pages/data_format.html
    download.file(x, sprintf("%s/gtf.gz", tempdir()), mode="wb")
    gtf <- read_tsv(sprintf("%s/gtf.gz", tempdir()), comment="#",
        col_names = c("chrm", "source", "feature_type", "start", "end",
            "score_not_used", "strand", "cds_phase", "additional"),
        col_types=cols(
            start=col_integer(),
            end=col_integer(),
            .default=col_character()))
    gtf
}

read_GENCODE_gtf_transcript <- function(gtf) {
    g <- gtf[gtf$feature_type == "transcript", ]
    g$transcript_id <- str_match(g$additional, 'transcript_id "([^"]*)"')[,2]
    g$transcript_name <- str_match(
        g$additional, 'transcript_name "([^"]*)"')[,2]
    g$transcript_type <- str_match(
        g$additional, 'transcript_type "([^"]*)"')[,2]
    g$gene_id <- str_match(g$additional, 'gene_id "([^"]*)"')[,2]
    g$gene_name <- str_match(g$additional, 'gene_name "([^"]*)"')[,2]
    g$gene_type <- str_match(g$additional, 'gene_type "([^"]*)"')[,2]
    ## there is also transcript_name, is it useful? not included
    g$level <- str_match(g$additional, 'level ([^;]*)')[,2]
    ## gene_status and transcript_status are obsolete after 25 and M11
    g
}

read_GENCODE_gtf_exon <- function(gtf) {
    g <- gtf[gtf$feature_type == "exon", ]
    g$transcript_id <- str_match(g$additional, 'transcript_id "([^"]*)"')[,2]
    g$exon_id <- str_match(g$additional, 'exon_id "([^"]*)"')[,2]
    g$exon_number <- str_match(g$additional, 'exon_number ([^;]*)')[,2]
    g
}

guess_chrmorder <- function(chrms) {
    chrms1 <- chrms[!(chrms %in% c("chrX","chrY","chrM"))]
    paste0("chr",c(as.character(seq_len(max(as.integer(str_replace(
        sort(unique(chrms1)), "chr", "")), na.rm=TRUE))), c("X","Y","M")))
}

#' build GENCODE gtf
#'
#' @param x GENCODE ftp url
#' @return GRangesList
#' @importFrom readr read_tsv
#' @importFrom readr cols
#' @importFrom readr col_integer
#' @importFrom readr col_character
#' @importFrom Seqinfo Seqinfo
#' @importFrom IRanges IRanges
#' @import stringr
#' @import GenomicRanges
build_GENCODE_gtf <- function(x) {

    gtf <- read_GENCODE_gtf(x)

    ## transcript
    g1 <- read_GENCODE_gtf_transcript(gtf)
    stopifnot(length(g1$transcript_id) == length(unique(g1$transcript_id)))

    ## exon
    g2 <- read_GENCODE_gtf_exon(gtf)
    chrms <- guess_chrmorder(g2$chrm)
    gr <- GRanges(seqnames = g2$chrm, ranges=IRanges(g2$start, g2$end),
        strand = g2$strand, seqinfo=Seqinfo(chrms))
    mcols(gr)$exon_number <- as.integer(g2$exon_number)
    names(gr) <- g2$exon_id
    grl <- GRangesList(split(gr, g2$transcript_id)) # slow
    stopifnot(length(grl) == length(g1$transcript_id))
    stopifnot(all(sort(names(grl)) == sort(g1$transcript_id)))

    ## CDS
    g3 <- gtf[gtf$feature_type == "CDS", ]
    g3$transcript_id <- str_match(g3$additional, 'transcript_id "([^"]*)"')[,2]
    tid2start <- vapply(split(g3$start, g3$transcript_id), min, numeric(1))
    tid2end <- vapply(split(g3$end, g3$transcript_id), max, numeric(1))
    g1$cdsStart <- tid2start[g1$transcript_id]
    g1$cdsEnd <- tid2end[g1$transcript_id]
    
    ## put together
    g1 <- g1[order(factor(g1$chrm, levels=chrms), g1$start),]
    grl <- grl[g1$transcript_id]
    mcl <- g1[match(names(grl), g1$transcript_id), c(
        "chrm", "start", "end", "strand",
        "transcript_id", "transcript_type", "transcript_name",
        "gene_name", "gene_id", "gene_type", "source", "level",
        "cdsStart", "cdsEnd")]
    colnames(mcl)[2] <- "transcript_start"
    colnames(mcl)[3] <- "transcript_end"
    colnames(mcl)[4] <- "transcript_strand"
    mcols(grl) <- mcl
    grl
}

#' convert transcript GRanges to gene GRanges
#'
#' @param txns GRanges object
#' @return a GRanges object
#' @examples
#' txns <- sesameData_getTxnGRanges("mm10")
#' genes <- sesameData_txnToGeneGRanges(txns)
#' 
#' @export
sesameData_txnToGeneGRanges <- function(txns) {
    gene_ids <- unique(txns$gene_id)
    gene2starts <- split(start(txns), txns$gene_id)[gene_ids]
    gene2ends <- split(end(txns), txns$gene_id)[gene_ids]
    genes <- GRanges(seqnames = seqnames(txns)[match(gene_ids, txns$gene_id)],
        IRanges(
            vapply(gene2starts, min, integer(1)),
            vapply(gene2ends, max, integer(1))),
        strand = strand(txns)[match(gene_ids, txns$gene_id)])

    names(genes) <- gene_ids
    mcols(genes)$gene_name <- txns$gene_name[match(names(genes), txns$gene_id)]
    mcols(genes)$gene_type <- txns$gene_type[match(names(genes), txns$gene_id)]
    sort(genes, ignore.strand=TRUE)
}

#' convert GRangesList to transcript GRanges
#'
#' @param genome hg38, mm10, ...
#' @param grl GRangesList object
#' @param merge2gene merge transcript to genes
#' @return a GRanges object
#' @examples
#'
#' ## all mm10 transcripts
#' txns <- sesameData_getTxnGRanges("mm10")
#' 
#' ## verified protein-coding transcripts
#' txns[(txns$transcript_type == "protein_coding" & txns$level <= 2)]
#'
#' ## merged to genes
#' sesameData_getTxnGRanges("mm10", merge2gene = TRUE)
#' 
#' @export
sesameData_getTxnGRanges <- function(
    genome = NULL, grl = NULL, merge2gene = FALSE) {
    
    if (is.null(grl)) {
        genome <- sesameData_check_genome(genome, NULL)
        grl <- sesameData_getGenomeInfo(genome)$txns
    }
    mcl <- mcols(grl)
    gr <- GRanges(
        seqnames = mcl$chrm, ranges = IRanges(
            mcl$transcript_start, mcl$transcript_end),
        strand = mcl$transcript_strand, seqinfo = seqinfo(grl))
    names(gr) <- mcl$transcript_id
    mcols(gr) <- mcl[,colnames(mcl)[!(colnames(mcl) %in% c(
        "chrm","transcript_id", "transcript_start",
        "transcript_end","transcript_strand"))]]

    gr <- sort(gr, ignore.strand = TRUE)
    if (merge2gene) {
        gr <- sesameData_txnToGeneGRanges(gr)
    }
    gr
}

---
title: "SeSAMe Data User Guide"
shorttitle: "sesameData guide"
package: sesameData
output: rmarkdown::html_vignette
fig_width: 6
fig_height: 5
vignette: >
    %\VignetteEngine{knitr::rmarkdown}
    %\VignetteIndexEntry{sesameData User Guide}
    %\VignetteEncoding{UTF-8}
---

# Introduction

`sesameData` package provides associated data for sesame package. This includes
example data for testing and instructional purpose, as we ll as probe annotation
for different Infinium platforms.

```{r, message=FALSE, warning=FALSE}
library(sesameData)
library(GenomicRanges)
sesameDataCache(c("genomeInfo.mm10", "HM450.address"))
```

# Genome Information

Sesame provides some utility functions to process transcript models, which can
be represented as data.frame, GRanges and GRangesList objects. For example,
`sesameData_getTxnGRanges` calls `sesameDataGet("genomeInfo.mm10")$txns` to
retrieve a transcript-centric GRangesList object from GENCODE including its
gene annotation, exon and cds (for protein-coding genes). It is then turned
into a simple GRanges object of transcript:

```{r}
sesameData_getTxnGRanges("mm10")
```

The returned GRanges object does not contain the exon coordinates. We can
further collapse different transcripts of the same gene (isoforms) to gene
level. Gene start is the minimum of all isoform starts and end is the maximum
of all isoform ends.

```{r}
sesameData_getTxnGRanges("mm10", merge2gene=TRUE)
```

# Probes


```{r message = FALSE, warning = FALSE}
library(GenomicRanges)
```

The following code get probes from different parts of the genome.
```{r eval=FALSE}
## probes in a region
sesameData_getProbesByRegion(GRanges('chr5',
    IRanges(135313937, 135419936)), platform = 'Mammal40')
## get chrX probes
sesameData_getProbesByRegion(chrm = 'chrX', platform = "Mammal40")
## get autosomal probes
sesameData_getProbesByRegion(
    chrm_to_exclude = c("chrX", "chrY"), platform = "Mammal40")
## get DNMT3A probes
sesameData_getProbesByGene('DNMT3A', "Mammal40", upstream=500)
## get DNMT3A promoter probes
sesameData_getProbesByGene('DNMT3A', "Mammal40", promoter = TRUE)
## get all promoter probes
sesameData_getProbesByGene(NULL, "Mammal40", promoter = TRUE)
```

One can annotate given probe ID using any genomic features stored in GRanges
objects. For example, the following demonstrate the annotation of 500 random
Mammal40 probes for gene promoters.
```{r eval=FALSE}
input_probes <- names(sesameData_getManifestGRanges("Mammal40"))[1:500]

## annotate for promoter
regs <- promoters(sesameData_getTxnGRanges("hg38"))
sesameData_annoProbes(input_probes, regs, column = "gene_name")

## annotate for gene association
regs <- sesameData_getTxnGRanges("hg38", merge2gene = TRUE)
sesameData_annoProbes(input_probes, regs, column = "gene_name")

## get genes associated with probes
regs <- sesameData_getTxnGRanges("hg38", merge2gene = TRUE)
sesameData_annoProbes(input_probes, regs, return_ov_features=TRUE)

## get genes associated with probes extending 10kb
input_probes <- c("cg14620903","cg22464003")
sesameData_annoProbes(input_probes, regs+10000, column = "gene_name")
```

# Manifest

Sesame provides access to array manfiest stored as GRanges object. These
GRanges object are converted from the raw tsv files on our [array annotation
website](http://zwdzwd.github.io/InfiniumAnnotation) using the [conversion
code](https://tinyurl.com/8bt9dssc)

```{r}
gr <- sesameData_getManifestGRanges("HM450")
length(gr)
```

Note that by default the GRanges object exclude decoy sequence probes (e.g.,
_alt, and _random contigs). To include them, we need to use the `decoy = TRUE`
option in sesameData_getManifestDF.

# Raw Data Retrieval

Titles of all the available data can be shown with:
```{r}
head(sesameDataList())
```

Each sesame datum from ExperimentHub is accessible through the `sesameDataGet`
interface. It should be noted that all data must be pre-cached to local disk
before they can be used.  This design is to prevent conflict in annotation data
caching and remove internet dependency. Caching needs only be done once per
sesame/sesameData installation. One can cache data using

```{r}
sesameDataCache()
```

Once a data object is loaded, it is stored to a tempoary cache, so that the data
doesn't need to be retrieved again next time we call `sesameDataGet`. This
design is meant to speeed up the run time.

For example, the annotation for HM27 can be retrieved with the title: 
```{r}
HM27.address <- sesameDataGet('HM27.address')
```

It's worth noting that once a data is retrieved through the `sesameDataGet`
inferface (below), it will stay in memory so next time the object will be
returned immediately. This design avoids repeated disk/web retrieval. In some
rare situation, one may want to redo the download/disk IO, or empty the cache
to save memory. This can be done with:

```{r}
sesameDataGet_resetEnv()
```


```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sesameDataGet.R
\name{sesameDataHas}
\alias{sesameDataHas}
\title{Whether sesameData has}
\usage{
sesameDataHas(data_titles)
}
\arguments{
\item{data_titles}{data titles to check}
}
\value{
a boolean vector the same length as data_titles
}
\description{
Whether sesameData has
}
\examples{
sesameDataHas(c("EPIC.address","EPIC.address.Nonexist"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sesameDataGet.R
\name{sesameDataList}
\alias{sesameDataList}
\title{List all SeSAMe data}
\usage{
sesameDataList(filter = NULL, full = FALSE)
}
\arguments{
\item{filter}{keyword to filter title, optional}

\item{full}{whether to display all columns}
}
\value{
all titles from SeSAMe Data
}
\description{
List all SeSAMe data
}
\examples{
sesameDataList("KYCG")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sesameDataCache.R
\name{sesameDataCacheAll}
\alias{sesameDataCacheAll}
\title{Cache all SeSAMe data}
\usage{
sesameDataCacheAll(data_titles = NULL)
}
\arguments{
\item{data_titles}{data to cache, if not given will cache all}
}
\value{
TRUE
}
\description{
Cache all SeSAMe data
}
\examples{
sesameDataCache("genomeInfo.hg38")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sesameDataCache.R
\name{sesameDataCache}
\alias{sesameDataCache}
\title{Cache SeSAMe data}
\usage{
sesameDataCache(data_titles = NULL)
}
\arguments{
\item{data_titles}{data to cache, if not given will cache all}
}
\value{
TRUE
}
\description{
Cache SeSAMe data
}
\examples{
sesameDataCache("genomeInfo.hg38")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/genome.R
\name{sesameData_getTxnGRanges}
\alias{sesameData_getTxnGRanges}
\title{convert GRangesList to transcript GRanges}
\usage{
sesameData_getTxnGRanges(genome = NULL, grl = NULL, merge2gene = FALSE)
}
\arguments{
\item{genome}{hg38, mm10, ...}

\item{grl}{GRangesList object}

\item{merge2gene}{merge transcript to genes}
}
\value{
a GRanges object
}
\description{
convert GRangesList to transcript GRanges
}
\examples{

## all mm10 transcripts
txns <- sesameData_getTxnGRanges("mm10")

## verified protein-coding transcripts
txns[(txns$transcript_type == "protein_coding" & txns$level <= 2)]

## merged to genes
sesameData_getTxnGRanges("mm10", merge2gene = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sesameDataGet.R
\name{sesameDataGet}
\alias{sesameDataGet}
\title{Get SeSAMe data}
\usage{
sesameDataGet(title, verbose = FALSE)
}
\arguments{
\item{title}{title of the data}

\item{verbose}{whether to output ExperimentHub message}
}
\value{
data object
}
\description{
Get SeSAMe data
}
\examples{

sesameDataCache("EPIC.1.SigDF")
EPIC.1.SigDF <- sesameDataGet('EPIC.1.SigDF')
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/globals.R
\docType{data}
\name{df_master}
\alias{df_master}
\title{Master data frame for all object to cache}
\format{
A data frame with 22 columns:
\describe{
  \item{Comments}{Additional comments}
  \item{EHID}{ExperimentHub ID}
  \item{VERSION}{sesameData version}
  \item{IN_USE}{Logical indicating if the resource is in use}
  \item{Title}{Title of the data resource}
  \item{Description}{Description of the data resource}
  \item{BiocVersion}{Bioconductor version}
  \item{Genome}{Genome build (e.g., hg38, mm10)}
  \item{SourceType}{Source file type}
  \item{SourceUrl}{URL to source}
  \item{SourceVersion}{Version of source data}
  \item{Species}{Species name}
  \item{TaxonomyId}{NCBI Taxonomy ID}
  \item{Coordinate_1_based}{Logical indicating if coordinates are 1-based}
  \item{DataProvider}{Data provider name}
  \item{Maintainer}{Maintainer contact information}
  \item{RDataClass}{R data class}
  \item{DispatchClass}{Dispatch class for ExperimentHub}
  \item{RDataPath}{Path to RData file}
  \item{Location_Prefix}{URL prefix for data location}
  \item{Tags}{Tags for categorization}
  \item{Notes}{Additional notes}
}
}
\value{
master sheet of sesameData objects
}
\description{
This is an internal object which will be updated on every new release
library(ExperimentHub)
eh <- query(ExperimentHub(localHub=FALSE), c("sesameData", "v1.13.1"))
data.frame(name=eh$title, eh=names(eh))
}
\details{
Cache location is default to
/Users/zhouw3/Library/Caches/org.R-project.R/R/ExperimentHub/
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{extend}
\alias{extend}
\title{Extend a GRanges}
\usage{
extend(gr, upstream = 0, downstream = 0)
}
\arguments{
\item{gr}{a GenomicRanges::GRanges}

\item{upstream}{distance to expand upstream}

\item{downstream}{distance to expand downstream}
}
\value{
a GenomicRanges::GRanges
}
\description{
source: https://support.bioconductor.org/p/78652/
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/platform.R
\name{inferPlatformFromProbeIDs}
\alias{inferPlatformFromProbeIDs}
\title{infer platform from Probe_IDs}
\usage{
inferPlatformFromProbeIDs(Probe_IDs, silent = FALSE)
}
\arguments{
\item{Probe_IDs}{probe IDs}

\item{silent}{suppress message}
}
\value{
a platform code
}
\description{
infer platform from Probe_IDs
}
\examples{
sesameDataCache("probeIDSignature")
inferPlatformFromProbeIDs(c("cg14620903","cg22464003"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/globals.R
\name{sesameDataGet_checkEnv}
\alias{sesameDataGet_checkEnv}
\title{Check whether the title exists in cacheEnv}
\usage{
sesameDataGet_checkEnv(title)
}
\arguments{
\item{title}{the title to check}
}
\value{
the data associated with the title or NULL if title doesn't exist
}
\description{
Check whether the title exists in cacheEnv
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/genome.R
\name{sesameData_txnToGeneGRanges}
\alias{sesameData_txnToGeneGRanges}
\title{convert transcript GRanges to gene GRanges}
\usage{
sesameData_txnToGeneGRanges(txns)
}
\arguments{
\item{txns}{GRanges object}
}
\value{
a GRanges object
}
\description{
convert transcript GRanges to gene GRanges
}
\examples{
txns <- sesameData_getTxnGRanges("mm10")
genes <- sesameData_txnToGeneGRanges(txns)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/genome.R
\name{sesameData_getGenomeInfo}
\alias{sesameData_getGenomeInfo}
\title{Get genome info files}
\usage{
sesameData_getGenomeInfo(genome)
}
\arguments{
\item{genome}{hg38, mm10, or GRanges with a metadata(genome)[["genome"]]}
}
\value{
a list of genome info files
}
\description{
Get genome info files
}
\examples{
sesameDataCache("genomeInfo.hg38")
res <- sesameData_getGenomeInfo("hg38")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/genome.R
\name{sesameData_check_genome}
\alias{sesameData_check_genome}
\title{Find genome assembly version(s) supported for a platform}
\usage{
sesameData_check_genome(genome, platform)
}
\arguments{
\item{genome}{mm10, hg38, ..., or NULL}

\item{platform}{HM27, HM450, EPIC, EPICv2, MSA, ...}
}
\value{
genome as string
}
\description{
Find genome assembly version(s) supported for a platform
}
\examples{
sesameData_check_genome(NULL, "Mammal40")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sesameData_getProbes.R
\name{sesameData_getProbesByGene}
\alias{sesameData_getProbesByGene}
\title{Get Probes by Genes or Gene Promoters}
\usage{
sesameData_getProbesByGene(
  gene_name = NULL,
  platform = NULL,
  promoter = FALSE,
  upstream = 1500,
  downstream = 1500,
  genome = NULL
)
}
\arguments{
\item{gene_name}{gene name, if NULL return all genes}

\item{platform}{EPIC or HM450}

\item{promoter}{if TRUE, use TSS instead of the whole gene}

\item{upstream}{number of bases to expand upstream of target gene}

\item{downstream}{number of bases to expand downstream of target gene}

\item{genome}{hg38 or hg19}
}
\value{
GRanges containing probes that fall into the given gene
}
\description{
Get probes mapped to a gene. All transcripts for the gene are considered.
The function takes a gene name as appears in UCSC RefGene database. The
platform and reference genome build can be changed with `platform` and
`genome` options. The function returns a vector of probes that falls
into the given gene.
}
\examples{

## download needed data
sesameDataCache(c("Mammal40.address", "genomeInfo.hg38"))

## get all probes overlapping with DNMT3A
probes <- sesameData_getProbesByGene(
    'DNMT3A', "Mammal40", upstream=500, downstream=500)

## get the promoter-associated probes
probes <- sesameData_getProbesByGene('DNMT3A', "Mammal40", promoter = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/platform.R
\name{sesameData_check_platform}
\alias{sesameData_check_platform}
\title{Check platform code}
\usage{
sesameData_check_platform(platform = NULL, probes = NULL, silent = TRUE)
}
\arguments{
\item{platform}{input platform}

\item{probes}{probes by which the platform may be guessed}

\item{silent}{suppress message}
}
\value{
platform code
}
\description{
Note: custome platforms lead to error here.
}
\examples{
sesameData_check_platform("HM450")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/genome.R
\name{build_GENCODE_gtf}
\alias{build_GENCODE_gtf}
\title{build GENCODE gtf}
\usage{
build_GENCODE_gtf(x)
}
\arguments{
\item{x}{GENCODE ftp url}
}
\value{
GRangesList
}
\description{
build GENCODE gtf
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sesameData_annoProbes.R
\name{sesameData_annoProbes}
\alias{sesameData_annoProbes}
\title{Annotate Probes by Probe ID}
\usage{
sesameData_annoProbes(
  Probe_IDs,
  regs = NULL,
  collapse = TRUE,
  chooseOne = FALSE,
  column = NULL,
  sep = ",",
  return_ov_probes = FALSE,
  return_ov_features = FALSE,
  out_name = NULL,
  platform = NULL,
  genome = NULL,
  silent = FALSE
)
}
\arguments{
\item{Probe_IDs}{a character vector of probe IDs}

\item{regs}{a GenomicRanges::GRanges object against which
probes will be annotated, default to genes if not given}

\item{collapse}{whether to collapse multiple regs into one}

\item{chooseOne}{choose an arbitrary annotation if multiple exist
default to FALSE. which concatenates all with ","}

\item{column}{which column in regs to annotate, if not given
return all overlapping probes}

\item{sep}{the delimiter for collapsing}

\item{return_ov_probes}{if TRUE will return overlapping
probes in a GRanges object.}

\item{return_ov_features}{if TRUE will return overlapping
features in a GRanges object.}

\item{out_name}{column header of the annotation, use column if not given}

\item{platform}{EPIC, MM285 etc. will infer from Probe_IDs if not given}

\item{genome}{hg38, mm10, ... will infer if not given.
For additional mapping, download the GRanges object from
http://zwdzwd.github.io/InfiniumAnnotation
and provide the following argument
..., genome = sesameAnno_buildManifestGRanges("downloaded_file"),...
to this function.}

\item{silent}{suppress messages}
}
\value{
a GRanges with annotated column
If a probe has no overlap with regs, it will be included in the results
with NA. But if a probe is not included in the manifest (due to
mappability), it won't be included in the results.
}
\description{
This function annotates probes based on genomic coordinate overlaps with
provided genomic regions (GRanges). Columns in the manifests will be added
to the annotation. Please note that if unfound, the annotation will be NA.
The probe will always be kept in the output.
}
\details{
For annotation by probe ID using KYCG databases (rather than by genomic
coordinates), see \code{knowYourCG::annoProbes()}
}
\examples{
library(GenomicRanges)
sesameDataCache(c(
    "genomeInfo.mm10", "MM285.address",
    "genomeInfo.hg38", "Mammal40.address"))

## Example 1: Basic usage - annotate with gene names (default)
## When regs=NULL, function defaults to gene annotation
probes <- c("cg14620903", "cg22464003")
anno <- sesameData_annoProbes(probes)
## Returns GRanges with gene_name column

## Example 2: Annotate mouse probes with promoter regions
regs <- sesameData_getTxnGRanges("mm10")
Probe_IDs <- names(sesameData_getManifestGRanges("MM285"))[1:100]
anno <- sesameData_annoProbes(Probe_IDs, promoters(regs), column="gene_name")
## Probes overlapping promoters are annotated with gene names

## Example 3: Get features that overlap with probes
## return_ov_features=TRUE returns the overlapping genomic features
genes <- sesameData_getTxnGRanges("hg38", merge2gene = TRUE)
ov_genes <- sesameData_annoProbes(
    c("cg14620903","cg22464003"), genes, return_ov_features=TRUE)
## Returns GRanges of genes overlapping the probes

## Example 4: Get probes that overlap with features
## return_ov_probes=TRUE returns only overlapping probes
ov_probes <- sesameData_annoProbes(
    c("cg14620903","cg22464003"), genes, return_ov_probes=TRUE)
## Returns GRanges of probes that overlap genes

## Example 5: Control annotation collapsing behavior
## collapse=TRUE (default): multiple annotations concatenated with separator
anno_collapsed <- sesameData_annoProbes(
    c("cg14620903","cg22464003"), genes, column="gene_name",
    collapse=TRUE, sep=";")
## Multiple overlapping genes listed as "GENE1;GENE2;GENE3"

## collapse=FALSE: each probe-feature overlap gets separate entry
anno_expanded <- sesameData_annoProbes(
    c("cg14620903","cg22464003"), genes, column="gene_name",
    collapse=FALSE)
## Probes with multiple overlaps appear multiple times

## Example 6: Choose only first annotation when multiple exist
anno_one <- sesameData_annoProbes(
    c("cg14620903","cg22464003"), genes, column="gene_name",
    chooseOne=TRUE)
## Each probe gets only the first overlapping gene

## Example 7: Annotate with custom genomic regions
custom_regs <- GRanges(
    seqnames = c("chr5", "chr5"),
    ranges = IRanges(start = c(10000, 135350870),
                     end = c(135350866, 145369531)),
    feature_type = c("enhancer", "silencer"))
anno_custom <- sesameData_annoProbes(
    c("cg14620903","cg22464003"), custom_regs,
    column="feature_type", genome="hg38")

## Note: For annotation by probe ID using KYCG databases
## (rather than genomic coordinates), see knowYourCG::annoProbes()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/globals.R
\name{sesameDataGet_resetEnv}
\alias{sesameDataGet_resetEnv}
\title{Empty cache environment to free memory}
\usage{
sesameDataGet_resetEnv()
}
\value{
gc() output
}
\description{
When this function is called sesameDataGet will
retrieve all data from disk again instead of using the in-memory
cache, i.e., sesameData:::cacheEnv.
}
\details{
Note this is different from sesameDataClearHub which empties the
ExperimentHub on disk.
}
\examples{
sesameDataGet_resetEnv()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sesameData_getProbes.R
\name{sesameData_getProbesByRegion}
\alias{sesameData_getProbesByRegion}
\title{Get probes by genomic region}
\usage{
sesameData_getProbesByRegion(
  regs,
  chrm = NULL,
  beg = 1,
  end = -1,
  platform = NULL,
  chrm_to_exclude = NULL,
  genome = NULL
)
}
\arguments{
\item{regs}{GRanges}

\item{chrm}{chromosome, when given regs are ignored}

\item{beg}{begin, 1 if omitted}

\item{end}{end, chromosome end if omitted}

\item{platform}{EPICv2, EPIC, HM450, ...}

\item{chrm_to_exclude}{chromosome to exclude.}

\item{genome}{hg38, mm10, ... will infer if not given.
For additional mapping, download the GRanges object from
http://zwdzwd.github.io/InfiniumAnnotation
and provide the following argument
..., genome = sesameAnno_buildManifestGRanges("downloaded_file"),...
to this function.}
}
\value{
GRanges of selected probes
}
\description{
The function takes a genomic coordinate and output the a vector of probes
on the specified platform that falls in the given genomic region.
}
\examples{

## download needed data
sesameDataCache(c("Mammal40.address", "genomeInfo.hg38"))

## get probes in a region
library(GenomicRanges)
probes = sesameData_getProbesByRegion(
    GRanges('chr5', IRanges(135313937, 135419936)), platform = 'Mammal40')

## get all probes on chromosome 5
probes = sesameData_getProbesByRegion(chrm = "chr5", platform = "Mammal40")

## get all probes on chromosome X
probes = sesameData_getProbesByRegion(chrm = 'chrX', platform = "Mammal40")

## get all probes on both chromosome X and Y
probes = sesameData_getProbesByRegion(
    chrm = c('chrX', 'chrY'), platform = "Mammal40")

## get all autosomal probes
probes = sesameData_getProbesByRegion(
    chrm_to_exclude = c("chrX", "chrY"), platform = "Mammal40")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sesameData_getManifestGRanges.R
\name{sesameData_getManifestGRanges}
\alias{sesameData_getManifestGRanges}
\title{get Infinium manifest GRanges}
\usage{
sesameData_getManifestGRanges(platform, genome = NULL)
}
\arguments{
\item{platform}{Mammal40, MM285, EPIC, and HM450}

\item{genome}{hg38, mm10, ... will infer if not given.
For additional mapping, download the GRanges object from
http://zwdzwd.github.io/InfiniumAnnotation
and provide the following argument
..., genome = sesameAnno_buildManifestGRanges("downloaded_file"),...
to this function.}
}
\value{
GRanges
}
\description{
Note that some unaligned probes are not included.
For full manifest, please visit
\url{http://zwdzwd.github.io/InfiniumAnnotation}
}
\examples{
sesameDataCache("Mammal40.address")
res <- sesameData_getManifestGRanges("Mammal40")
}
